/*
	MIT License http://www.opensource.org/licenses/mit-license.php
	Author Tobias Koppers @sokra
*/

"use strict";

const HotModuleReplacementPlugin = require("../HotModuleReplacementPlugin");
const InnerGraph = require("../optimize/InnerGraph");
const ConstDependency = require("./ConstDependency");
const HarmonyAcceptDependency = require("./HarmonyAcceptDependency");
const HarmonyAcceptImportDependency = require("./HarmonyAcceptImportDependency");
const HarmonyEvaluatedImportSpecifierDependency = require("./HarmonyEvaluatedImportSpecifierDependency");
const HarmonyExportPresenceImportSpecifierDependency = require("./HarmonyExportPresenceImportSpecifierDependency");
const HarmonyExports = require("./HarmonyExports");
const { ExportPresenceModes } = require("./HarmonyImportDependency");
const HarmonyImportSideEffectDependency = require("./HarmonyImportSideEffectDependency");
const HarmonyImportSpecifierDependency = require("./HarmonyImportSpecifierDependency");

/** @typedef {import("estree").ExportAllDeclaration} ExportAllDeclaration */
/** @typedef {import("estree").ExportNamedDeclaration} ExportNamedDeclaration */
/** @typedef {import("estree").Identifier} Identifier */
/** @typedef {import("estree").ImportDeclaration} ImportDeclaration */
/** @typedef {import("estree").ImportExpression} ImportExpression */
/** @typedef {import("../../declarations/WebpackOptions").JavascriptParserOptions} JavascriptParserOptions */
/** @typedef {import("../javascript/JavascriptParser")} JavascriptParser */
/** @typedef {import("../optimize/InnerGraph").InnerGraph} InnerGraph */
/** @typedef {import("../optimize/InnerGraph").TopLevelSymbol} TopLevelSymbol */
/** @typedef {import("./HarmonyImportDependency")} HarmonyImportDependency */

const harmonySpecifierTag = Symbol("harmony import");

/**
 * @typedef {Object} HarmonySettings
 * @property {string[]} ids
 * @property {string} source
 * @property {number} sourceOrder
 * @property {string} name
 * @property {boolean} await
 * @property {Record<string, any> | undefined} assertions
 */

/**
 * @param {ImportDeclaration | ExportNamedDeclaration | ExportAllDeclaration | ImportExpression} node node with assertions
 * @returns {Record<string, any> | undefined} assertions
 */
function getAssertions(node) {
	// TODO remove cast when @types/estree has been updated to import assertions
	const assertions = /** @type {{ assertions?: ImportAttributeNode[] }} */ (
		node
	).assertions;
	if (assertions === undefined) {
		return undefined;
	}
	const result = {};
	for (const assertion of assertions) {
		const key =
			assertion.key.type === "Identifier"
				? assertion.key.name
				: assertion.key.value;
		result[key] = assertion.value.value;
	}
	return result;
}

module.exports = class HarmonyImportDependencyParserPlugin {
	/**
	 * @param {JavascriptParserOptions} options options
	 */
	constructor(options) {
		this.exportPresenceMode =
			options.importExportsPresence !== undefined
				? ExportPresenceModes.fromUserOption(options.importExportsPresence)
				: options.exportsPresence !== undefined
				? ExportPresenceModes.fromUserOption(options.exportsPresence)
				: options.strictExportPresence
				? ExportPresenceModes.ERROR
				: ExportPresenceModes.AUTO;
		this.strictThisContextOnImports = options.strictThisContextOnImports;
	}

	/**
	 * @param {JavascriptParser} parser the parser
	 * @returns {void}
	 */
	apply(parser) {
		const { exportPresenceMode } = this;

		function getNonOptionalPart(members, membersOptionals) {
			let i = 0;
			while (i < members.length && membersOptionals[i] === false) i++;
			return i !== members.length ? members.slice(0, i) : members;
		}

		function getNonOptionalMemberChain(node, count) {
			while (count--) node = node.object;
			return node;
		}

		/**
		 * @param {string[]} ids ids
		 * @returns {string} guard name
		 */
		const createGuard = ids => ids.join(".");
		/**
		 * @param {string} guard guard
		 * @param {number} idsLength ids length
		 * @returns {number} mode
		 */
		const detectExportPresenceMode = (guard, idsLength) => {
			if (
				exportPresenceMode === ExportPresenceModes.NONE ||
				// namespace objects are safe to use
				idsLength === 0 ||
				parser.scope.guards.has(guard) ||
				// if possible guard is in guard position,
				// it is guarded by member chain minus one element. e.g.
				// if (a && a.b) {} or if (a && "c" in a.b) {}
				// a.b is guarded by a
				(parser.scope.inGuardPosition &&
					(idsLength === 1 ||
						parser.scope.guards.has(guard.slice(0, guard.lastIndexOf(".")))))
			)
				return ExportPresenceModes.NONE;

			return exportPresenceMode;
		};
		parser.hooks.isPure
			.for("Identifier")
			.tap("HarmonyImportDependencyParserPlugin", expression => {
				const expr = /** @type {Identifier} */ (expression);
				if (
					parser.isVariableDefined(expr.name) ||
					parser.getTagData(expr.name, harmonySpecifierTag)
				) {
					return true;
				}
			});
		parser.hooks.import.tap(
			"HarmonyImportDependencyParserPlugin",
			(statement, source) => {
				parser.state.lastHarmonyImportOrder =
					(parser.state.lastHarmonyImportOrder || 0) + 1;
				const clearDep = new ConstDependency(
					parser.isAsiPosition(statement.range[0]) ? ";" : "",
					statement.range
				);
				clearDep.loc = statement.loc;
				parser.state.module.addPresentationalDependency(clearDep);
				parser.unsetAsiPosition(statement.range[1]);
				const assertions = getAssertions(statement);
				const sideEffectDep = new HarmonyImportSideEffectDependency(
					source,
					parser.state.lastHarmonyImportOrder,
					assertions
				);
				sideEffectDep.loc = statement.loc;
				parser.state.module.addDependency(sideEffectDep);
				return true;
			}
		);
		parser.hooks.importSpecifier.tap(
			"HarmonyImportDependencyParserPlugin",
			(statement, source, id, name) => {
				const ids = id === null ? [] : [id];
				parser.tagVariable(name, harmonySpecifierTag, {
					name,
					source,
					ids,
					sourceOrder: parser.state.lastHarmonyImportOrder,
					assertions: getAssertions(statement)
				});
				return true;
			}
		);
		parser.hooks.binaryExpression.tap(
			"HarmonyImportDependencyParserPlugin",
			expression => {
				if (expression.operator === "in") {
					const leftPartEvaluated = parser.evaluateExpression(expression.left);
					if (!leftPartEvaluated || leftPartEvaluated.couldHaveSideEffects())
						return;
					const leftPart = leftPartEvaluated.asString();
					if (!leftPart) return;

					const rightPart = parser.evaluateExpression(expression.right);
					if (!rightPart || !rightPart.isIdentifier()) return;

					const rootInfo = rightPart.rootInfo;
					if (
						!rootInfo ||
						!rootInfo.tagInfo ||
						rootInfo.tagInfo.tag !== harmonySpecifierTag
					)
						return;
					const settings = rootInfo.tagInfo.data;
					const members = rightPart.getMembers();
					const baseIds = settings.ids.concat(members);
					const ids = baseIds.concat([leftPart]);
					const dep = new HarmonyEvaluatedImportSpecifierDependency(
						settings.source,
						settings.sourceOrder,
						ids,
						settings.name,
						expression.range,
						settings.assertions,
						"in"
					);
					dep.directImport = members.length === 0;
					dep.asiSafe = !parser.isAsiPosition(expression.range[0]);
					dep.loc = expression.loc;
					parser.state.module.addDependency(dep);
					InnerGraph.onUsage(parser.state, e => (dep.usedByExports = e));

					if (parser.scope.inGuardPosition) {
						parser.scope.guards.add(createGuard(ids));
						// namespace objects and identifiers are safe to use
						if (baseIds.length > 1) {
							// check for export presence for right side expression
							const mode = detectExportPresenceMode(
								createGuard(baseIds),
								baseIds.length
							);
							// e.g if ("a" in b.c) {}
							// here "b.c" is not guarded by "b"
							if (mode !== ExportPresenceModes.NONE) {
								const dep = new HarmonyExportPresenceImportSpecifierDependency(
									settings.source,
									settings.sourceOrder,
									baseIds,
									settings.name,
									rightPart.expression.range,
									mode,
									settings.assertions
								);
								dep.loc = rightPart.expression.loc;
								parser.state.module.addDependency(dep);
							}
						}
					}

					return true;
				} else if (expression.operator === "!=") {
					if (!parser.scope.inGuardPosition) return;
					let identifierEvaluated;
					const leftPartEvaluated = parser.evaluateExpression(expression.left);
					if (!leftPartEvaluated) {
						return;
					} else if (leftPartEvaluated.isIdentifier()) {
						const rootInfo = leftPartEvaluated.rootInfo;
						if (
							!rootInfo ||
							!rootInfo.tagInfo ||
							rootInfo.tagInfo.tag !== harmonySpecifierTag
						)
							return;
						identifierEvaluated = leftPartEvaluated;
					} else if (leftPartEvaluated.isFalsy() !== true) {
						return;
					}

					const rightPartEvaluated = parser.evaluateExpression(
						expression.right
					);
					if (!rightPartEvaluated) {
						return;
					} else if (
						!identifierEvaluated &&
						rightPartEvaluated.isIdentifier()
					) {
						const rootInfo = rightPartEvaluated.rootInfo;
						if (
							!rootInfo ||
							!rootInfo.tagInfo ||
							rootInfo.tagInfo.tag !== harmonySpecifierTag
						)
							return;
						identifierEvaluated = rightPartEvaluated;
					} else if (rightPartEvaluated.isFalsy() !== true) {
						return;
					}

					// other hooks will add guards and dependencies
					parser.walkExpression(identifierEvaluated.expression);
					return true;
				}
			}
		);
		parser.hooks.expression
			.for(harmonySpecifierTag)
			.tap("HarmonyImportDependencyParserPlugin", expr => {
				const settings = /** @type {HarmonySettings} */ (parser.currentTagData);
				let exportPresenceModeComputed;
				// namespace object is safe to use
				if (settings.ids.length) {
					const guard = createGuard(settings.ids);
					exportPresenceModeComputed = detectExportPresenceMode(
						guard,
						settings.ids.length
					);
					if (parser.scope.inGuardPosition) parser.scope.guards.add(guard);
				}
				const dep = new HarmonyImportSpecifierDependency(
					settings.source,
					settings.sourceOrder,
					settings.ids,
					settings.name,
					expr.range,
					exportPresenceModeComputed || ExportPresenceModes.NONE,
					settings.assertions
				);
				dep.shorthand = parser.scope.inShorthand;
				dep.directImport = true;
				dep.asiSafe = !parser.isAsiPosition(expr.range[0]);
				dep.loc = expr.loc;
				parser.state.module.addDependency(dep);
				InnerGraph.onUsage(parser.state, e => (dep.usedByExports = e));
				return true;
			});
		parser.hooks.expressionMemberChain
			.for(harmonySpecifierTag)
			.tap(
				"HarmonyImportDependencyParserPlugin",
				(expression, members, membersOptionals) => {
					const settings = /** @type {HarmonySettings} */ (
						parser.currentTagData
					);
					const nonOptionalMembers = getNonOptionalPart(
						members,
						membersOptionals
					);
					const expr =
						nonOptionalMembers !== members
							? getNonOptionalMemberChain(
									expression,
									members.length - nonOptionalMembers.length
							  )
							: expression;
					const ids = settings.ids.concat(nonOptionalMembers);
					const guard = createGuard(ids);
					const exportPresenceMode = detectExportPresenceMode(
						guard,
						ids.length
					);
					if (parser.scope.inGuardPosition) parser.scope.guards.add(guard);
					const dep = new HarmonyImportSpecifierDependency(
						settings.source,
						settings.sourceOrder,
						ids,
						settings.name,
						expr.range,
						exportPresenceMode,
						settings.assertions
					);
					dep.asiSafe = !parser.isAsiPosition(expr.range[0]);
					dep.loc = expr.loc;
					parser.state.module.addDependency(dep);
					InnerGraph.onUsage(parser.state, e => (dep.usedByExports = e));
					return true;
				}
			);
		parser.hooks.callMemberChain
			.for(harmonySpecifierTag)
			.tap(
				"HarmonyImportDependencyParserPlugin",
				(expression, members, membersOptionals) => {
					const { arguments: args, callee } = expression;
					const settings = /** @type {HarmonySettings} */ (
						parser.currentTagData
					);
					const nonOptionalMembers = getNonOptionalPart(
						members,
						membersOptionals
					);
					const expr =
						nonOptionalMembers !== members
							? getNonOptionalMemberChain(
									callee,
									members.length - nonOptionalMembers.length
							  )
							: callee;
					const ids = settings.ids.concat(nonOptionalMembers);
					const dep = new HarmonyImportSpecifierDependency(
						settings.source,
						settings.sourceOrder,
						ids,
						settings.name,
						expr.range,
						detectExportPresenceMode(createGuard(ids), ids.length),
						settings.assertions
					);
					dep.directImport = members.length === 0;
					dep.call = true;
					dep.asiSafe = !parser.isAsiPosition(expr.range[0]);
					// only in case when we strictly follow the spec we need a special case here
					dep.namespaceObjectAsContext =
						members.length > 0 && this.strictThisContextOnImports;
					dep.loc = expr.loc;
					parser.state.module.addDependency(dep);
					if (args) parser.walkExpressions(args);
					InnerGraph.onUsage(parser.state, e => (dep.usedByExports = e));
					return true;
				}
			);
		const { hotAcceptCallback, hotAcceptWithoutCallback } =
			HotModuleReplacementPlugin.getParserHooks(parser);
		hotAcceptCallback.tap(
			"HarmonyImportDependencyParserPlugin",
			(expr, requests) => {
				if (!HarmonyExports.isEnabled(parser.state)) {
					// This is not a harmony module, skip it
					return;
				}
				const dependencies = requests.map(request => {
					const dep = new HarmonyAcceptImportDependency(request);
					dep.loc = expr.loc;
					parser.state.module.addDependency(dep);
					return dep;
				});
				if (dependencies.length > 0) {
					const dep = new HarmonyAcceptDependency(
						expr.range,
						dependencies,
						true
					);
					dep.loc = expr.loc;
					parser.state.module.addDependency(dep);
				}
			}
		);
		hotAcceptWithoutCallback.tap(
			"HarmonyImportDependencyParserPlugin",
			(expr, requests) => {
				if (!HarmonyExports.isEnabled(parser.state)) {
					// This is not a harmony module, skip it
					return;
				}
				const dependencies = requests.map(request => {
					const dep = new HarmonyAcceptImportDependency(request);
					dep.loc = expr.loc;
					parser.state.module.addDependency(dep);
					return dep;
				});
				if (dependencies.length > 0) {
					const dep = new HarmonyAcceptDependency(
						expr.range,
						dependencies,
						false
					);
					dep.loc = expr.loc;
					parser.state.module.addDependency(dep);
				}
			}
		);
	}
};

module.exports.harmonySpecifierTag = harmonySpecifierTag;
module.exports.getAssertions = getAssertions;
